@@ -5,6 +5,7 @@ import sys
 import logging
 
 import numpy as np
+import pyarrow as pa
 
 import dask.base
 from vaex.expression import Expression
@@ -19,6 +20,8 @@ from .docstrings import docsubst
 import vaex.utils
 
 
+list_ = list # dangerous, we override list in this module (the aggregator)
+
 logger = logging.getLogger("vaex.agg")
 
 if vaex.utils.has_c_extension:
@@ -49,14 +52,18 @@ class aggregation_encoding:
         if type == '_sum_moment':
             if 'parameters' in agg_spec:  # renameing between spec and implementation
                 agg_spec['moment'] = agg_spec.pop('parameters')[0]
-        if type == 'first':
-            args = agg_spec.pop('expression')
+        if 'expressions' in agg_spec:
+            args = agg_spec.pop('expressions')
+        if type == "list":
+            if "parameters" in agg_spec:  # renameing between spec and implementation
+                agg_spec["dropnan"], agg_spec["dropmissing"] = agg_spec.pop("parameters")
         return f(*args, **agg_spec)
 
 
-class AggregatorDescriptor(object):
+class AggregatorDescriptor:
     def __repr__(self):
-        return 'vaex.agg.{}({!r})'.format(self.short_name, str(self.expression))
+        args = [*self.expressions]
+        return 'vaex.agg.{}({!r})'.format(self.short_name, ", ".join(map(str, args)))
 
     def pretty_name(self, id, df):
         if id is None:
@@ -75,7 +82,6 @@ class AggregatorExpressionUnary(AggregatorDescriptor):
         self.code = code
         self.expressions = self.agg.expressions
         self.selection = self.agg.selection
-        self.expression = self.agg.expression
 
     def __repr__(self):
         return f'{self.code}{self.agg!r}'
@@ -214,49 +220,44 @@ for op in _unary_ops:
 
 
 class AggregatorDescriptorBasic(AggregatorDescriptor):
-    def __init__(self, name, expression, short_name, multi_args=False, agg_args=[], selection=None, edges=False):
+    def __init__(self, name, expressions, short_name, multi_args=False, agg_args=[], selection=None, edges=False):
         self.name = name
         self.short_name = short_name
-        self.expression = str(expression)
         self.agg_args = agg_args
         self.edges = edges
         self.selection = _normalize_selection_name(selection)
-        if not multi_args:
-            if self.expression == '*':
-                self.expressions = []
-            else:
-                self.expressions = [self.expression]
-        else:
-            self.expressions = expression
+        assert isinstance(expressions, (list_, tuple))
+        for e in expressions:
+            assert not isinstance(e, (list_, tuple))
+        self.expressions = [str(k) for k in expressions]
+        if len(self.expressions) == 1 and self.expressions[0] == '*':
+            self.expressions = []
 
     def __repr__(self):
-        if self.agg_args:
-            return 'vaex.agg.{}({!r}, {})'.format(self.short_name, str(self.expression), ", ".join(map(str, self.agg_args)))
-        else:
-            return 'vaex.agg.{}({!r})'.format(self.short_name, str(self.expression))
+        args = [*self.expressions, *self.agg_args]
+        return 'vaex.agg.{}({!r})'.format(self.short_name, ", ".join(map(str, args)))
 
     def encode(self, encoding):
         spec = {'aggregation': self.short_name}
         if len(self.expressions) == 0:
             pass
-        elif len(self.expressions) == 1:
-            spec['expression'] = self.expression
         else:
-            spec['expression'] = [str(k) for k in self.expressions]
+            spec['expressions'] = [str(k) for k in self.expressions]
         if self.selection is not None:
             spec['selection'] = str(self.selection) if isinstance(self.selection, Expression) else self.selection
         if self.edges:
             spec['edges'] = True
-        if self.agg_args:
+        if self.agg_args and self.short_name not in ['first', 'last']:
             spec['parameters'] = self.agg_args
         return spec
 
     def _prepare_types(self, df):
-        if self.expression == '*':
+        if len(self.expressions) == 0 and self.short_name == "count":
             self.dtype_in = DataType(np.dtype('int64'))
             self.dtype_out = DataType(np.dtype('int64'))
         else:
-            self.dtype_in = df[str(self.expressions[0])].data_type().index_type
+            self.dtypes_in = [df[str(e)].data_type().index_type for e in self.expressions]
+            self.dtype_in = self.dtypes_in[0]
             self.dtype_out = self.dtype_in
             if self.short_name == "count":
                 self.dtype_out = DataType(np.dtype('int64'))
@@ -275,8 +276,18 @@ class AggregatorDescriptorBasic(AggregatorDescriptor):
         return [task], finish(task)
 
     def _create_operation(self, grid, nthreads):
-        agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtype_in)
-        bytes_per_cell = self.dtype_out.numpy.itemsize
+        if self.name in ["AggFirst", "AggList"]:
+            if len(self.dtypes_in) == 1:
+                # rows use int64
+                agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtypes_in[0], vaex.dtype(np.dtype('int64')))
+            else:
+                agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtypes_in[0], self.dtypes_in[1])
+        else:
+            agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtype_in)
+        if self.dtype_out.is_primitive or self.dtype_out.is_temporal:
+            bytes_per_cell = self.dtype_out.numpy.itemsize
+        else:
+            bytes_per_cell = self.dtype_out.value_type.numpy.itemsize
         cells = reduce(operator.mul, [len(binner) for binner in grid.binners], 1)
         grids = nthreads
         ncells = len(grid)
@@ -292,12 +303,21 @@ class AggregatorDescriptorBasic(AggregatorDescriptor):
             grids = 1
         if logger.isEnabledFor(logging.INFO):
             logger.info("Using %r grids for %r thread for aggerator %r (total grid cells %s)", grids, nthreads, self, f"{ncells:,}")
-        predicted_memory_usage = bytes_per_cell * cells * grids
-        vaex.memory.local.agg.pre_alloc(predicted_memory_usage, f"aggregator data for {agg_op_type}")
+        if self.short_name in ["list"]:
+            # cannot predict memory usage
+            predicted_memory_usage = None
+            grids = 1
+        else:
+            predicted_memory_usage = bytes_per_cell * cells * grids
+            vaex.memory.local.agg.pre_alloc(predicted_memory_usage, f"aggregator data for {agg_op_type}")
         agg_op = agg_op_type(grid, grids, nthreads, *self.agg_args)
         used_memory = sys.getsizeof(agg_op)
-        if predicted_memory_usage != used_memory:
-            raise RuntimeError(f'Wrong prediction for {agg_op_type}, expected to take {predicted_memory_usage} bytes but actually used {used_memory}')
+
+        if predicted_memory_usage is not None:
+            if predicted_memory_usage != used_memory:
+                raise RuntimeError(f'Wrong prediction for {agg_op_type}, expected to take {predicted_memory_usage} bytes but actually used {used_memory}')
+        else:
+            vaex.memory.local.agg.pre_alloc(used_memory, f"aggregator data for {agg_op_type}")
         return agg_op
 
     def get_result(self, agg_operation):
@@ -351,22 +371,26 @@ class AggregatorDescriptorNUnique(AggregatorDescriptorBasic):
 
 class AggregatorDescriptorMulti(AggregatorDescriptor):
     """Uses multiple operations/aggregation to calculate the final aggretation"""
-    def __init__(self, name, expression, short_name, selection=None, edges=False):
+    def __init__(self, name, expressions, short_name, selection=None, edges=False):
         self.name = name
         self.short_name = short_name
-        self.expression = str(expression)
-        self.expressions = [self.expression]
+        self.expressions = expressions
         self.selection = selection
         self.edges = edges
+        assert isinstance(expressions, (list_, tuple))
+        for e in expressions:
+            assert not isinstance(e, (list_, tuple))
+        self.expressions = [str(k) for k in expressions]
 
 
 class AggregatorDescriptorMean(AggregatorDescriptorMulti):
-    def __init__(self, name, expression, short_name="mean", selection=None, edges=False):
-        super(AggregatorDescriptorMean, self).__init__(name, expression, short_name, selection=selection, edges=edges)
+    def __init__(self, name, expressions, short_name="mean", selection=None, edges=False):
+        super(AggregatorDescriptorMean, self).__init__(name, expressions, short_name, selection=selection, edges=edges)
+        assert len(expressions) == 1
 
     def add_tasks(self, df, binners, progress):
         progressbar = vaex.utils.progressbars(progress, title=repr(self))
-        expression = expression_sum = expression = df[str(self.expression)]
+        expression = expression_sum = expression = df[str(self.expressions[0])]
 
         sum_agg = sum(expression_sum, selection=self.selection, edges=self.edges)
         count_agg = count(expression, selection=self.selection, edges=self.edges)
@@ -401,7 +425,7 @@ class AggregatorDescriptorVar(AggregatorDescriptorMulti):
 
     def add_tasks(self, df, binners, progress):
         progressbar = vaex.utils.progressbars(progress, title=repr(self))
-        expression_sum = expression = df[str(self.expression)]
+        expression_sum = expression = df[str(self.expressions[0])]
         expression = expression_sum = expression.astype('float64')
         sum_moment = _sum_moment(str(expression_sum), 2, selection=self.selection, edges=self.edges)
         sum_ = sum(str(expression_sum), selection=self.selection, edges=self.edges)
@@ -431,6 +455,69 @@ class AggregatorDescriptorVar(AggregatorDescriptorMulti):
         return [task_sum_moment, task_sum, task_count], finish(task_sum_moment, task_sum, task_count)
 
 
+class AggregatorDescriptorSkew(AggregatorDescriptorMulti):
+    def __init__(self, name, expression, short_name='skew', selection=None, edges=False):
+        super(AggregatorDescriptorSkew, self).__init__(name, [expression], short_name, selection=selection, edges=edges)
+
+    def add_tasks(self, df, binners, progress):
+        progressbar = vaex.utils.progressbars(progress, title=repr(self))
+        expression = expression_sum = expression = df[str(self.expressions[0])]
+        expression = expression_sum = expression.astype('float64')
+
+        sum_moment1 = _sum_moment(str(expression_sum), 1, selection=self.selection, edges=self.edges)
+        sum_moment2 = _sum_moment(str(expression_sum), 2, selection=self.selection, edges=self.edges)
+        sum_moment3 = _sum_moment(str(expression_sum), 3, selection=self.selection, edges=self.edges)
+        count_ = count(str(expression), selection=self.selection, edges=self.edges)
+
+        task_sum_moment1 = sum_moment1.add_tasks(df, binners, progress=progressbar)[0][0]
+        task_sum_moment2 = sum_moment2.add_tasks(df, binners, progress=progressbar)[0][0]
+        task_sum_moment3 = sum_moment3.add_tasks(df, binners, progress=progressbar)[0][0]
+        task_count = count_.add_tasks(df, binners, progress=progressbar)[0][0]
+
+        @vaex.delayed
+        def finish(sum_moment1, sum_moment2, sum_moment3, count):
+            with np.errstate(divide='ignore', invalid='ignore'):
+                m1 = sum_moment1 / count
+                m2 = sum_moment2 / count
+                m3 = sum_moment3 / count
+                skew = (m3 - 3*m1*m2 + 2*m1**3) / (m2 - m1**2)**(3/2)
+            return self.finish(skew)
+        return [task_sum_moment1, task_sum_moment2, task_sum_moment3, task_count], finish(task_sum_moment1, task_sum_moment2, task_sum_moment3, task_count)
+
+
+class AggregatorDescriptorKurtosis(AggregatorDescriptorMulti):
+    def __init__(self, name, expression, short_name='kurtosis', selection=None, edges=False):
+        super(AggregatorDescriptorKurtosis, self).__init__(name, [expression], short_name, selection=selection, edges=edges)
+
+    def add_tasks(self, df, binners, progress):
+        progressbar = vaex.utils.progressbars(progress, title=repr(self))
+        expression = expression_sum = expression = df[str(self.expressions[0])]
+        expression = expression_sum = expression.astype('float64')
+
+        sum_moment1 = _sum_moment(str(expression_sum), 1, selection=self.selection, edges=self.edges)
+        sum_moment2 = _sum_moment(str(expression_sum), 2, selection=self.selection, edges=self.edges)
+        sum_moment3 = _sum_moment(str(expression_sum), 3, selection=self.selection, edges=self.edges)
+        sum_moment4 = _sum_moment(str(expression_sum), 4, selection=self.selection, edges=self.edges)
+        count_ = count(str(expression), selection=self.selection, edges=self.edges)
+
+        task_sum_moment1 = sum_moment1.add_tasks(df, binners, progress=progressbar)[0][0]
+        task_sum_moment2 = sum_moment2.add_tasks(df, binners, progress=progressbar)[0][0]
+        task_sum_moment3 = sum_moment3.add_tasks(df, binners, progress=progressbar)[0][0]
+        task_sum_moment4 = sum_moment4.add_tasks(df, binners, progress=progressbar)[0][0]
+        task_count = count_.add_tasks(df, binners, progress=progressbar)[0][0]
+
+        @vaex.delayed
+        def finish(sum_moment1, sum_moment2, sum_moment3, sum_moment4, count):
+            with np.errstate(divide='ignore', invalid='ignore'):
+                m1 = sum_moment1 / count
+                m2 = sum_moment2 / count
+                m3 = sum_moment3 / count
+                m4 = sum_moment4 / count
+                kurtosis = (m4 - 4*m1*m3 + 6*m1**2*m2 - 3*m1**4) / (m2 - m1**2)**2 -3.0
+            return self.finish(kurtosis)
+        return [task_sum_moment1, task_sum_moment2, task_sum_moment3, task_sum_moment4, task_count], finish(task_sum_moment1, task_sum_moment2, task_sum_moment3, task_sum_moment4, task_count)
+
+
 class AggregatorDescriptorStd(AggregatorDescriptorVar):
     def finish(self, value):
         return value**0.5
@@ -438,62 +525,91 @@ class AggregatorDescriptorStd(AggregatorDescriptorVar):
 @register
 def count(expression='*', selection=None, edges=False):
     '''Creates a count aggregation'''
-    return AggregatorDescriptorBasic('AggCount', expression, 'count', selection=selection, edges=edges)
+    return AggregatorDescriptorBasic('AggCount', [expression], 'count', selection=selection, edges=edges)
 
 @register
 def sum(expression, selection=None, edges=False):
     '''Creates a sum aggregation'''
-    return AggregatorDescriptorBasic('AggSum', expression, 'sum', selection=selection, edges=edges)
+    return AggregatorDescriptorBasic('AggSum', [expression], 'sum', selection=selection, edges=edges)
 
 @register
 def mean(expression, selection=None, edges=False):
     '''Creates a mean aggregation'''
-    return AggregatorDescriptorMean('mean', expression, 'mean', selection=selection, edges=edges)
+    return AggregatorDescriptorMean('mean', [expression], 'mean', selection=selection, edges=edges)
 
 @register
 def min(expression, selection=None, edges=False):
     '''Creates a min aggregation'''
-    return AggregatorDescriptorBasic('AggMin', expression, 'min', selection=selection, edges=edges)
+    return AggregatorDescriptorBasic('AggMin', [expression], 'min', selection=selection, edges=edges)
 
 @register
 def _sum_moment(expression, moment, selection=None, edges=False):
     '''Creates a sum of moment aggregator'''
-    return AggregatorDescriptorBasic('AggSumMoment', expression, '_sum_moment', agg_args=[moment], selection=selection, edges=edges)
+    return AggregatorDescriptorBasic('AggSumMoment', [expression], '_sum_moment', agg_args=[moment], selection=selection, edges=edges)
 
 @register
 def max(expression, selection=None, edges=False):
     '''Creates a max aggregation'''
-    return AggregatorDescriptorBasic('AggMax', expression, 'max', selection=selection, edges=edges)
+    return AggregatorDescriptorBasic('AggMax', [expression], 'max', selection=selection, edges=edges)
 
 @register
-def first(expression, order_expression, selection=None, edges=False):
-    '''Creates a first aggregation'''
-    return AggregatorDescriptorBasic('AggFirst', [expression, order_expression], 'first', multi_args=True, selection=selection, edges=edges)
+def first(expression, order_expression=None, selection=None, edges=False):
+    '''Creates a first aggregation.
+
+    :param expression: {expression_one}.
+    :param order_expression:  Order the values in the bins by this expression.
+    :param selection: {selection1}
+    :param edges: {edges}
+    '''
+    return AggregatorDescriptorBasic('AggFirst', [expression, order_expression] if order_expression is not None else [expression], 'first', multi_args=True, selection=selection, edges=edges, agg_args=[False])
+
+@register
+@docsubst
+def last(expression, order_expression=None, selection=None, edges=False):
+    '''Creates a first aggregation.
+
+    :param expression: {expression_one}.
+    :param order_expression:  Order the values in the bins by this expression.
+    :param selection: {selection1}
+    :param edges: {edges}
+    '''
+    return AggregatorDescriptorBasic('AggFirst', [expression, order_expression] if order_expression is not None else [expression], 'last', multi_args=True, selection=selection, edges=edges, agg_args=[True])
 
 @register
 def std(expression, ddof=0, selection=None, edges=False):
     '''Creates a standard deviation aggregation'''
-    return AggregatorDescriptorStd('std', expression, 'std', ddof=ddof, selection=selection, edges=edges)
+    return AggregatorDescriptorStd('std', [expression], 'std', ddof=ddof, selection=selection, edges=edges)
 
 @register
 def var(expression, ddof=0, selection=None, edges=False):
     '''Creates a variance aggregation'''
-    return AggregatorDescriptorVar('var', expression, 'var', ddof=ddof, selection=selection, edges=edges)
+    return AggregatorDescriptorVar('var', [expression], 'var', ddof=ddof, selection=selection, edges=edges)
 
 @register
+def skew(expression, selection=None, edges=False):
+    '''Create a skew aggregation.'''
+    return AggregatorDescriptorSkew('skew', expression, 'skew', selection=selection, edges=edges)
+
+@register
+def kurtosis(expression, selection=None, edges=False):
+    '''Create a kurtosis aggregation.'''
+    return AggregatorDescriptorKurtosis('kurtosis', expression, 'kurtosis', selection=selection, edges=edges)
+
+@register
+@docsubst
 def nunique(expression, dropna=False, dropnan=False, dropmissing=False, selection=None, edges=False):
     """Aggregator that calculates the number of unique items per bin.
 
     :param expression: {expression_one}
-    :param dropmissing: do not count missing values
-    :param dropnan: do not count nan values
-    :param dropna: short for any of the above, (see :func:`Expression.isna`)
+    :param dropmissing: {dropmissing}
+    :param dropnan: {dropnan}
+    :param dropna: {dropna}
     :param selection: {selection1}
     """
     if dropna:
         dropnan = True
         dropmissing = True
-    return AggregatorDescriptorNUnique('AggNUnique', expression, 'nunique', dropmissing, dropnan, selection=selection, edges=edges)
+    return AggregatorDescriptorNUnique('AggNUnique', [expression], 'nunique', dropmissing, dropnan, selection=selection, edges=edges)
 
 @docsubst
 def any(expression=None, selection=None):
@@ -535,6 +651,29 @@ def all(expression=None, selection=None):
                 return sum(f'astype({expression}, "bool") & astype({selection}, "bool")') == count(expression)
 
 
+@register
+@docsubst
+class list(AggregatorDescriptorBasic):
+    '''Aggregator that returns a list of values belonging to the specified expression.
+
+    :param expression: {expression_one}
+    :param selection: {selection1}
+    :param dropmissing: {dropmissing}
+    :param dropnan: {dropnan}
+    :param dropna: {dropna}
+    :param edges: {edges}
+    '''
+    def __init__(self, expression, selection=None, dropna=False, dropnan=False, dropmissing=False, edges=False):
+        if dropna:
+            dropnan = True
+            dropmissing = True
+        super(list, self).__init__("AggList", [expression], "list", selection=selection, edges=edges, agg_args=[dropnan, dropmissing])
+
+    def _prepare_types(self, df):
+        super()._prepare_types(df)
+        self.dtype_out = vaex.dtype(pa.large_list(self.dtype_out.arrow))
+
+
 # @register
 # def covar(x, y):
 #     '''Creates a standard deviation aggregation'''
