@@ -2,6 +2,8 @@ import ast
 import copy
 import os
 import base64
+import datetime
+from pydoc import doc
 import time
 import cloudpickle as pickle
 import functools
@@ -130,12 +132,19 @@ class Meta(type):
                             assert b.ds == a.ds
                             b = b.expression
                         elif isinstance(b, (np.timedelta64)):
-                            unit, step = np.datetime_data(b)
+                            unit, step = np.datetime_data(b.dtype)
                             assert step == 1
                             b = b.astype(np.uint64).item()
                             b = f'scalar_timedelta({b}, {unit!r})'
                         elif isinstance(b, (np.datetime64)):
                             b = f'scalar_datetime("{b}")'
+                        elif isinstance(b, np.ndarray) and b.ndim == 0 and vaex.dtype_of(b).is_datetime:
+                            b = f'scalar_datetime("{b}")'
+                        elif isinstance(b, np.ndarray) and b.ndim == 0 and vaex.dtype_of(b).is_timedelta:
+                            unit, step = np.datetime_data(b.dtype)
+                            assert step == 1
+                            b = b.astype(np.uint64).item()
+                            b = f'scalar_timedelta({b}, {unit!r})'
                         expression = '({0} {1} {2})'.format(a.expression, op['code'], b)
                     return Expression(self.ds, expression=expression)
                 attrs['__%s__' % op['name']] = f
@@ -597,7 +606,7 @@ class Expression(with_metaclass(Meta)):
             indices, fields = slicer
         else:
             raise NotImplementedError
-        
+
         if indices != slice(None):
             expr = self.df[indices][self.expression]
         else:
@@ -923,6 +932,20 @@ class Expression(with_metaclass(Meta)):
         kwargs['expression'] = self.expression
         return self.ds.var(**kwargs)
 
+    def skew(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
+        '''Shortcut for df.skew(expression, ...), see `DataFrame.skew`'''
+        kwargs = dict(locals())
+        del kwargs['self']
+        kwargs['expression'] = self.expression
+        return self.df.skew(**kwargs)
+
+    def kurtosis(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
+        '''Shortcut for df.kurtosis(expression, ...), see `DataFrame.kurtosis`'''
+        kwargs = dict(locals())
+        del kwargs['self']
+        kwargs['expression'] = self.expression
+        return self.df.kurtosis(**kwargs)
+
     def minmax(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
         '''Shortcut for ds.minmax(expression, ...), see `Dataset.minmax`'''
         kwargs = dict(locals())
@@ -958,6 +981,7 @@ class Expression(with_metaclass(Meta)):
         """Alias to df.is_masked(expression)"""
         return self.ds.is_masked(self.expression)
 
+    @docsubst
     def value_counts(self, dropna=False, dropnan=False, dropmissing=False, ascending=False, progress=False, axis=None):
         """Computes counts of unique values.
 
@@ -965,10 +989,11 @@ class Expression(with_metaclass(Meta)):
           * If the expression/column is not categorical, it will be converted on the fly
           * dropna is False by default, it is True by default in pandas
 
-        :param dropna: when True, it will not report the NA (see :func:`Expression.isna`)
-        :param dropnan: when True, it will not report the nans(see :func:`Expression.isnan`)
-        :param dropmissing: when True, it will not report the missing values (see :func:`Expression.ismissing`)
+        :param dropna: {dropna}
+        :param dropnan: {dropnan}
+        :param dropmissing: {dropmissing}
         :param ascending: when False (default) it will report the most frequent occuring item first
+        :param progress: {progress}
         :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
         :returns: Pandas series containing the counts
         """
@@ -993,10 +1018,15 @@ class Expression(with_metaclass(Meta)):
         counters = [None] * self.ds.executor.thread_pool.nthreads
         def map(thread_index, i1, i2, selection_masks, blocks):
             ar = blocks[0]
+            if len(ar) == 0:
+                return 0
             if counters[thread_index] is None:
                 counters[thread_index] = counter_type(1)
             if data_type.is_list and axis is None:
-                ar = ar.values
+                try:
+                    ar = ar.values
+                except AttributeError:  # pyarrow ChunkedArray
+                    ar = ar.combine_chunks().values
             if data_type_item.is_string:
                 ar = _to_string_sequence(ar)
             else:
@@ -1046,9 +1076,9 @@ class Expression(with_metaclass(Meta)):
             else:
                 null_offset = 0
             if dropmissing and counter.has_null:
-                deletes.append(null_offset)
+                deletes.append(counter.null_index)
             if dropnan and counter.has_nan:
-                deletes.append(0)
+                deletes.append(counter.nan_index)
             if vaex.array_types.is_arrow_array(keys):
                 indices = np.delete(np.arange(len(keys)), deletes)
                 keys = keys.take(indices)
@@ -1081,14 +1111,16 @@ class Expression(with_metaclass(Meta)):
     def unique(self, dropna=False, dropnan=False, dropmissing=False, selection=None, axis=None, limit=None, limit_raise=True, array_type='list', progress=None, delay=False):
         """Returns all unique values.
 
-        :param dropmissing: do not count missing values
-        :param dropnan: do not count nan values
-        :param dropna: short for any of the above, (see :func:`Expression.isna`)
+        :param dropna: {dropna}
+        :param dropnan: {dropnan}
+        :param dropmissing: {dropmissing}
+        :param selection: {selection}
         :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
         :param int limit: {limit}
         :param bool limit_raise: {limit_raise}
-        :param progress: {progress}
         :param bool array_type: {array_type}
+        :param progress: {progress}
+        :param bool delay: {delay}
         """
         return self.ds.unique(self, dropna=dropna, dropnan=dropnan, dropmissing=dropmissing, selection=selection, array_type=array_type, axis=axis, limit=limit, limit_raise=limit_raise, progress=progress, delay=delay)
 
@@ -1096,12 +1128,15 @@ class Expression(with_metaclass(Meta)):
     def nunique(self, dropna=False, dropnan=False, dropmissing=False, selection=None, axis=None, limit=None, limit_raise=True, progress=None, delay=False):
         """Counts number of unique values, i.e. `len(df.x.unique()) == df.x.nunique()`.
 
-        :param dropmissing: do not count missing values
-        :param dropnan: do not count nan values
-        :param dropna: short for any of the above, (see :func:`Expression.isna`)
+        :param dropna: {dropna}
+        :param dropnan: {dropnan}
+        :param dropmissing: {dropmissing}
+        :param selection: {selection}
+        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
         :param int limit: {limit}
         :param bool limit_raise: {limit_raise}
-        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
+        :param progress: {progress}
+        :param bool delay: {delay}
         """
         def key_function():
             fp = vaex.cache.fingerprint(self.fingerprint(), dropna, dropnan, dropmissing, selection, axis, limit)
