@@ -218,16 +218,18 @@ def ismissing(x):
             x = _to_string_sequence(x)
             mask = x.mask()
             if mask is None:
-                mask = np.zeros(x.length, dtype=np.bool)
+                mask = np.zeros(x.length, dtype=bool)
             return mask
         elif isinstance(x, np.ndarray) and x.dtype.kind in 'O':
             return x == None
         else:
-            return np.zeros(len(x), dtype=np.bool)
+            return np.zeros(len(x), dtype=bool)
 
 
 @register_function()
 def notmissing(x):
+    if isinstance(x, vaex.array_types.supported_arrow_array_types):
+        return pa.compute.invert(ismissing(x))
     return ~ismissing(x)
 
 
@@ -245,11 +247,13 @@ def isnan(x):
         else:
             return x != x
     else:
-        return np.zeros(len(x), dtype=np.bool)
+        return np.zeros(len(x), dtype=bool)
 
 
 @register_function()
 def notnan(x):
+    if isinstance(x, vaex.array_types.supported_arrow_array_types):
+        return pa.compute.invert(isnan(x))
     return ~isnan(x)
 
 
@@ -262,6 +266,8 @@ def isna(x):
 @register_function()
 def notna(x):
     """Opposite of isna"""
+    if isinstance(x, vaex.array_types.supported_arrow_array_types):
+        return pa.compute.invert(isna(x))
     return ~isna(x)
 
 
@@ -1735,6 +1741,13 @@ def str_replace(x, pat, repl, n=-1, flags=0, regex=False):
     """
     sl = _to_string_sequence(x).replace(pat, repl, n, flags, regex)
     return column.ColumnStringArrow.from_string_sequence(sl)
+    # TODO: we should be using arrow compute functions (since MB rewrote them in arrow)
+    # except that they don't support flags yet (case insensitive)
+    # we could invoke it when flags=0 though
+    # if regex:
+    #     return pc.replace_substring_regex(x, pattern=pat, max_replacements=None if n==-1 else n, replacement=repl)
+    # else:
+    #     return pc.replace_substring(x, pattern=pat, max_replacements=None if n==-1 else n, replacement=repl)
 
 
 @register_function(scope='str')
@@ -2513,8 +2526,13 @@ def _astype(x, dtype):
                     units = dtype[len('datetime64')+1:-1]
                 else:
                     units = 'ns'
-                dtype = pa.timestamp(units)
-
+                if vaex.dtype_of(x) == str and units in ['m', 'h', 'D', 'W', 'M', 'Y']:
+                    # it's a slower path, but it works
+                    # using pc.strptime seems to offset by a day orso?
+                    x = np.array(x, dtype=dtype)
+                    return x.astype(dtype)
+                else:
+                    dtype = pa.timestamp(units)
             else:  # parse dtype
                 if len(dtype) > len('timedelta64'):
                     units = dtype[len('timedelta64')+1:-1]
@@ -2522,6 +2540,9 @@ def _astype(x, dtype):
                     units = 'ns'
                 dtype = pa.duration(units)
 
+        if dtype == 'int':  # "int" is not automatically recognized by arrow, unlike "float" for example.
+            dtype = pa.int64()
+
         y = x.cast(dtype, safe=False)
         return y
     else:  # numpy case
